/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "inductance.h"

inductance::inductance() {
	nb_params=6;
	message="";
}

int inductance::setType(int choice) {
	int status;
	
	switch (choice) {
		case 0: /* // wires */
			type=ParalleleWires;
			break;
		case 1: /* strip line */
			type=StripLine;
			status=setTabKsh();
			break;
		case 2: /* converging line */
			type=ConvergingLine;
			status=setTabKsh();
			break;
		case 3: /* disks */
			type=Disks;
			break;
		case 4: /* hollow cylinder */
			type=HollowCylinder;
			status=setTabKw();
			break;
		case 5: /* coaxial line */
			type=CoaxialLine;
			break;
		case 6: /* thin solenoid */
			type=ThinSolenoid;
			status=setTabKl();
			break;
		case 7: /* torus */
			type=Torus;
			break;
		case 8: /* wire loop */
			type=WireLoop;
			break;
		case 9: /* rectangular coil */
			type=RectangularCoil;
			status=setTabKsh();
			status=setTabKl();
			break;
		case 10: /* helix */
			type=Helix;
			status=setTabKl();
			status=setTabG();
			status=setTabN();
			break;
		case 11: /* balls */
			type=Balls;
			break;
		default:
			return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;

}

double inductance::Ksh(double a) {
/* Calculates the coefficient for magnetic field leakage in a strip line */
	double K;
	int status;
;
	status=tab_Ksh.getValue(a,tabulation::InterpXLog);
	K=tab_Ksh.value;
	if(K>1.0) K=1.0;
	if(K<0.0) K=0.0;
	message=QString("Strip line field leakage correction factor: %1").arg(K,0,'g',2);
	if(status==EXIT_FAILURE) message="Error, calculation is wrong !";
	if(status==2) message.append("\nOut of table interpolation");
	printf("%s \n",message.ascii());
	
	return K;
}

double inductance::Kl(double a) {
/* Calculates the coefficient for magnetic field leakage in a strip line */
	double K;
	int status;

	status=tab_Kl.getValue(a,tabulation::InterpXLog);
	K=tab_Kl.value;
	if(K>1.0) K=1.0;
	if(K<0.0) K=0.0;
	message=QString("Solenoid field leakage correction factor: %1").arg(K,0,'g',2);
	if(status==EXIT_FAILURE) message="Error, calculation is wrong !";
	if(status==2) message.append("\nOut of table interpolation");
	printf("%s \n",message.ascii());
	
	return K;
}

double inductance::Ns(double a) {
/* Calculates the coefficient for magnetic field leakage in a strip line */
	double K;
	int status;
	
	status=tab_N.getValue(a,tabulation::InterpXLog);
	K=tab_N.value;
	if(K>0.325) K=0.325;
	if(K<0.0) K=0.0;
	message=QString("Helix N* correction factor: %1").arg(K,0,'g',2);
	if(status==EXIT_FAILURE) message="Error, calculation is wrong !";
	if(status==2) message.append("\nOut of table interpolation");
	printf("%s \n",message.ascii());
	
	return K;
}

double inductance::Gs(double a) {
/* Calculates the coefficient for magnetic field leakage in a strip line */
	double K;
	int status;
	
	status=tab_G.getValue(a,tabulation::InterpXLog);
	K=tab_G.value;
	if(K>0.51) K=0.51;
	if(K<-3) K=-3;
	message=QString("Helix G* correction factor: %1").arg(K,0,'g',2);
	if(status==EXIT_FAILURE) message="Error, calculation is wrong !";
	if(status==2) message.append("\nOut of table interpolation");
	printf("%s \n",message.ascii());
	
	return K;
}

double inductance::Kw(double a) {
/* Calculates the coefficient for magnetic field leakage in a strip line */
	double K;
	int status;
	
	status=tab_Kw.getValue(a,tabulation::InterpLinear);
	K=tab_Kw.value;
	message=QString("Hollow cylinder Kw correction factor: %1").arg(K,0,'g',2);
	if(status==EXIT_FAILURE) message="Error, calculation is wrong !";
	if(status==2) message.append("\nOut of table interpolation");
	printf("%s \n",message.ascii());
	
	return K;
}

double inductance::hollow_cylinder(void) {
	double length,r0,r1,K;
	
	r1=parameters[1];
	r0=parameters[2];
	length=parameters[3];
	K=Kw(r1/r0);
	L=MU_0/(2*PI)*length*(log(2*length/r0)-1+K);
	L_length=length;

	return L;
}

double inductance::wire_loop(void) {
	double l,a,d;
	
	d=parameters[1];
	a=parameters[2]/2;
	l=2*PI*a;
	if(a<(5*d)) message.append("Warning: out of model validity!\n");
	L=MU_0/(2*PI)*l*(log(4*l/d)-2.45);
	message.append("Check capacitance...\n");
	
	return L;
}

double inductance::thin_solenoid(void) {
	double a,length,N,K;
	
	a=parameters[1]/2;
	N=parameters[2];
	length=parameters[3];
	K=Kl(length/(2*a));
	L=MU_0*N*N*PI*a*a/length*K;
	message.append("Check capacitance...\n");

	return L;
}

double inductance::torus(void) {
	double a,N,R0;
	
	a=parameters[1]/2;
	N=parameters[2];
	R0=parameters[3];
	
	L=MU_0*N*N*(R0-sqrt( R0*R0-a*a ));
	message.append("Check capacitance...\n");

	return L;
}

double inductance::parallele_wires(void) {
	double length,gap,a,p;
	
	a=parameters[1]/2;
	gap=parameters[2];
	length=parameters[3];
	p=sqrt(0.25*gap*gap-a*a);
	L=-MU_0/PI*length*(log(gap/a)+log(0.5-p/gap));
	L_length=length;
	
	return L;
}

double inductance::coaxial_line(void) {
	double length,r1,r2;
	
	r1=parameters[1];
	r2=parameters[2];
	length=parameters[3];
	
	L=MU_0/(2.0*PI)*length*log(r2/r1);
	L_length=length;

	return L;
}

double inductance::strip_line(void) {
	double a,gap,width,length;

	width=parameters[1];
	gap=parameters[2];
	length=parameters[3];
	a=gap/width;
	
	L=MU_0*gap*length/width*Ksh(a);
	L_length=length;
	return L;
}

double inductance::converging_line(void) {
	double a,gap,w1,w2,length;
	double Kco;
	
	w2=parameters[1];
	gap=parameters[2];
	length=parameters[3];
	w1=parameters[4];
	a=gap/((w1+w2)/2.0);
	Kco=Ksh(a);
	
	L=MU_0*gap*length/(w1-w2)*log(w1/w2)*Kco;
	L_length=length;
	return L;
}

double inductance::disks(void) {
	double r1,r2,g1,g2;
	
	r1=parameters[1];
	r2=parameters[2];
	g1=parameters[3];
	g2=parameters[4];
	
	L=MU_0/(2.0*PI)*(g2-g1+ (g1*r2-g2*r1)/(r2-r1)*log(r2/r1) );
	message.append("Check capacitance...\n");
	return L;
}

double inductance::rectangular_coil(void) {
	double width,gap,length;
	double regime,c,Kr=-1;
	
	width=parameters[1];
	gap=parameters[2];
	length=parameters[3];
	
	regime=gap/length;
	if(regime<=0.2) Kr=Ksh(gap/width);
	if(regime>=0.3) Kr=Kl( width/(2*sqrt((gap*length)/PI)) );
	if(regime>0.2 && regime<0.3) {
	    c=(regime-0.2)/(0.3-0.2);
	    Kr=c*Kl( width/(2*sqrt((gap*length)/PI)) )+(1-c)*Ksh(gap/width);
	    message.append("\nWarning: between two models. Please, check results!");
	}
	L=MU_0*length*gap/width*Kr;
	L_length=length;
	
	return L;
}

double inductance::helix(void) {
	double radius,height,wire_radius,pitch;
	double Ls,kl,gs,ns,S,N;
	
	radius=parameters[1];
	height=parameters[2];
	wire_radius=parameters[3];
	pitch=parameters[4];
	
	S=PI*radius*radius;
	N=height/pitch;
	
	kl=Kl(height/(2*radius));
	Ls=MU_0*S/height*kl;
	gs=Gs(2*radius/pitch);
	ns=Ns(N);
	
	L=Ls-MU_0*radius*N*(gs+ns);
	message.append("Check capacitance...\n");
	
	return L;
}

double inductance::balls(void) {
	double r1,r2,d;
	
	r1=parameters[1]/2;
	r2=parameters[2]/2;
	d=parameters[3];
	
	if((d<(5*r1)) || (d<(5*r2))) message.append("Warning: out of model validity!\n");
	C=4*PI*E_0*( (1-r1*r2/(d*d)) / (1/r1+1/r2-2/d) );
	L=E_0*MU_0/C;
	message.append("Check inductance...\n");
	
	return L;
}

double inductance::L2C(void) {
	L=parameters[5];
	
	if(L>0) {
		C=E_0*MU_0/(L/L_length)*L_length;
	} else C=0;
	parameters[6]=C;
	
	return C;
}

double inductance::C2L(void) {
	C=parameters[6];
	
	if(C>0) {
		L=E_0*MU_0/(C/L_length)*L_length;
	} else L=0;
	parameters[5]=L;
	
	return L;
}

double inductance::f(void) {
//this function is a wrapper needed by the solver
	
	message="";
	L_length=1; //for calculating L from C or the opposite, we need an equivalent lenght
	switch(type) {
		case ParalleleWires:
			return parameters[5]-parallele_wires();
		case StripLine:
			return parameters[5]-strip_line();
		case ConvergingLine:
			return parameters[5]-converging_line();
		case Disks:
			return parameters[5]-disks();
		case HollowCylinder:
			return parameters[5]-hollow_cylinder();
		case CoaxialLine:
			return parameters[5]-coaxial_line();
		case ThinSolenoid:
			return parameters[5]-thin_solenoid();
		case Torus:
			return parameters[5]-torus();
		case WireLoop:
			return parameters[5]-wire_loop();
		case RectangularCoil:
			return parameters[5]-rectangular_coil();
		case Helix:
			return parameters[5]-helix();
		case Balls:
			return parameters[5]-balls();
	}
	return (double)EXIT_FAILURE;
}

void inductance::CalculateLandC(void) {
//we know that L and C have to be calculated -> direct evaluation
	
	parameters[5]=0; //because f() needs it to be 0
	L=-1*f();
	parameters[5]=L;
	parameters[6]=L2C();
}

int inductance::setTabKsh(void) {
	const short int Npts_Ksh=25;
	double X_Ksh[Npts_Ksh] = { 
	0.092, 0.107, 0.120, 0.139, 0.165, 0.192, 0.223, 0.253, 0.320, 0.367, 0.446, 0.517, 0.641, 0.749, 0.894, 1.130, 1.429, 1.897, 2.376, 3.155, 4.315, 5.511, 7.319, 8.644, 9.533 
	};
	double Y_Ksh[Npts_Ksh] = {
	0.876, 0.863, 0.851, 0.835, 0.813, 0.793, 0.772, 0.752, 0.715, 0.691, 0.656, 0.629, 0.584, 0.551, 0.514, 0.465, 0.413, 0.356, 0.313, 0.263, 0.214, 0.180, 0.147, 0.132, 0.124
	};
	
	return tab_Ksh.setTables(X_Ksh,Y_Ksh,Npts_Ksh);
}

int inductance::setTabKl(void) {
	const short int Npts_Kl=28;
	double X_Kl[Npts_Kl] = { 
	0.010, 0.018, 0.028, 0.046, 0.070, 0.093, 0.115, 0.144, 0.186, 0.262, 0.339, 0.446, 0.616, 0.865, 1.234, 1.572, 2.137, 3.201, 4.794, 6.208, 9.150, 13.484, 16.369, 21.544, 26.580, 40.459, 60.598, 93.741
	};
	double Y_Kl[Npts_Kl] = {
	0.043, 0.062, 0.085, 0.120, 0.165, 0.204, 0.239, 0.275, 0.323, 0.389, 0.441, 0.511, 0.586, 0.669, 0.744, 0.788, 0.836, 0.885, 0.920, 0.938, 0.956, 0.974, 0.979, 0.988, 0.993, 0.994, 0.999, 1.004
	};
	
	return tab_Kl.setTables(X_Kl,Y_Kl,Npts_Kl);
}

int inductance::setTabN(void) {
	const short int Npts_N=32;
	double X_N[Npts_N] = { 
	1.011, 1.095, 1.198, 1.324, 1.463, 1.633, 1.822, 1.974, 2.159, 2.385, 2.609, 2.854, 3.153, 3.733, 4.164, 4.785, 5.609, 6.573, 7.856, 9.205, 10.364, 13.011, 15.855, 19.318, 23.075, 27.022, 33.239, 40.890, 54.422, 71.009, 83.129, 101.239
	};
	double Y_N[Npts_N] = {
	0.008, 0.020, 0.034, 0.052, 0.068, 0.083, 0.098, 0.110, 0.122, 0.133, 0.145, 0.157, 0.169, 0.186, 0.197, 0.210, 0.224, 0.236, 0.248, 0.258, 0.264, 0.275, 0.284, 0.292, 0.298, 0.302, 0.307, 0.312, 0.317, 0.321, 0.322, 0.325
	};
	
	return tab_N.setTables(X_N,Y_N,Npts_N);
}

int inductance::setTabG(void) {
	const short int Npts_G=19;
	double X_G[Npts_G] = { 
	0.029, 0.033, 0.037, 0.046, 0.054, 0.064, 0.077, 0.095, 0.121, 0.154, 0.201, 0.267, 0.326, 0.405, 0.498, 0.595, 0.697, 0.832, 0.994
	};
	double Y_G[Npts_G] = {
	-2.969, -2.836, -2.694, -2.514, -2.343, -2.173, -2.002, -1.794, -1.567, -1.339, -1.074, -0.799, -0.600, -0.392, -0.183, -0.003, 0.158, 0.328, 0.508
	};
	
	return tab_G.setTables(X_G,Y_G,Npts_G);
}

int inductance::setTabKw(void) {
	const short int Npts_Kw=6;
	double X_Kw[Npts_Kw] = { 
	0.0, 0.2, 0.4, 0.6, 0.8, 1.0
	};
	double Y_Kw[Npts_Kw] = {
	0.250, 0.245, 0.188, 0.130, 0.066, 0.0
	};
	
	return tab_Kw.setTables(X_Kw,Y_Kw,Npts_Kw);
}

