/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "rlc.h"

double rlc::f(void) {
// the goal of that function is to be able to run Newton's method on
// these functions, if needed.
	switch (eqselection) {
		case fIsF0: 
			return F_0();
			break;
		case fIsF: 
			return F();
			break;
		case fIsU0: 
			return U_0();
			break;
		case fIsU: 
			return U();
			break;
	}
	return EXIT_FAILURE;
}

double rlc::F(void) {
/* calculation of the multiplier applied to sqrt(LC) */
	double f_factor;
	
	R=parameters[4];
	L=parameters[5];
	C=parameters[6];
	k=R/(sqrt(L/C));
	
	if(k<2) {
		f_factor=atan(sqrt(4.0/(k*k)-1))/(sqrt(1-k*k/4.0));
		return f_factor;
	}
	if(k==2){
		f_factor=1.0;
		return f_factor;
	}
	if(k>2){
		f_factor=log( (1+sqrt(1-4/(k*k))) / (1-sqrt(1-4/(k*k))) ) / k*sqrt(1-4/(k*k));
		return f_factor;	
	}
	return EXIT_FAILURE;
}

double rlc::U(void) {
/* calculation of the output voltage */
	double U;
	double f_factor;
	
	U0=parameters[3];
	R=parameters[4];
	L=parameters[5];
	C=parameters[6];
	k=R/(sqrt(L/C));
	
	if(k<2) {
		f_factor=F();
		U=U0*k/(sqrt(1-k*k/4))*exp(-k/2*f_factor)*sin(sqrt(1-k*k/4)*f_factor);
		return U;
	}
	if(k==2){
		U=U0*k/exp(1);
		return U;
	}
	if(k>2){
		U=k*U0*pow((1-sqrt(1-4/(k*k)))/(1+sqrt(1-4/(k*k))), 1/(2*sqrt(1-4/(k*k))));
		return U;	
	}
	return EXIT_FAILURE;
}

double rlc::F_0(void) {
/* calculation of f(k)-tmax/sqrt(LC) in order to look for its root */
	tmax=parameters[1];
	L=parameters[5];
	C=parameters[6];
	
	return F()-tmax/sqrt(L*C);
}

double rlc::U_0(void) {
/* calculation of U(k)-R*Imax in order to look for its root */
	Imax=parameters[2];
	R=parameters[4];
	
	return U()-R*Imax;
}


double rlc::dF_0(int index_variable) {
	/* calculates the derivative of F0 over variable given by its index (1->R, ...) */
	double var_org;
	double val1,val2;
	
	var_org=parameters[index_variable];
	parameters[index_variable]=var_org*(1+ALPHA);
	val1=F_0();
	parameters[index_variable]=var_org*(1-ALPHA);
	val2=F_0();
	parameters[index_variable]=var_org;
	
	return (val1-val2)/(2*ALPHA*var_org);
}

double rlc::dU_0(int index_variable) {
	/* calculates the derivative of F0 over variable given by its index (1->R, ...) */
	double var_org;
	double val1,val2;
	
	var_org=parameters[index_variable];
	parameters[index_variable]=var_org*(1+ALPHA);
	val1=U_0();
	parameters[index_variable]=var_org*(1-ALPHA);
	val2=U_0();
	parameters[index_variable]=var_org;
	
	return (val1-val2)/(2*ALPHA*var_org);
}


int rlc::solve_system(int index_to_solve1,int index_to_solve2,double guess1,double guess2) {
	/* this function solves the circuit equation, knowing Imax, tmax and 1 of RLC (given by index_known) */
	/* looking for the two other parameters */
	/* initial guesses must be provided, rounded BY EXCESS */
	double kx,ky;
	double df1_dx, df1_dy, df2_dx, df2_dy, f1, f2;
	int step=0;
	
	parameters[index_to_solve1]=guess1;
	parameters[index_to_solve2]=guess2;
	
	do {
		step++;
		/* step initialization */
		/* derivatives over x */
		df1_dx=dF_0(index_to_solve1);
		df2_dx=dU_0(index_to_solve1);
		/* derivatives over y */
		df1_dy=dF_0(index_to_solve2);
		df2_dy=dU_0(index_to_solve2);
		/* regular values */
		f1=F_0();
		f2=U_0();
		/* we calculate the increments */
		kx=(f2*df1_dy-f1*df2_dy)/(df1_dx*df2_dy-df2_dx*df1_dy);
		ky=(f2*df1_dx-f1*df2_dx)/(df1_dy*df2_dx-df2_dy*df1_dx);
		/* we add these increments to the variables */
		parameters[index_to_solve1] += kx;
		parameters[index_to_solve2] += ky;
		//values MUST be positives -> return a way to know who is not >0
		if(parameters[index_to_solve1] <= 0) {
			return -(1+10);
		}
		if(parameters[index_to_solve2] <= 0) {
			return -(2+10);
		}
		if(fabs(kx)<MINIMUM_CORRECTION || fabs(ky)<MINIMUM_CORRECTION){
			message="WARNING: equation solver is converging slowly\n";
			return EXIT_FAILURE;
		}
		if(step>MAXSTEPS) {
			message="ERROR: equation solver is not converging (too many steps)\n";
			return EXIT_FAILURE;
		}
	} while ((fabs(kx/parameters[index_to_solve1]) > TOLERANCE) || (fabs(ky/parameters[index_to_solve2]) > TOLERANCE));
	
	k=parameters[4]/sqrt(parameters[5]/parameters[6]);
	
	return EXIT_SUCCESS;
}

int rlc::setPhysicalGuesses(int index_to_solve[3],double guesses[3]) {
	int ind1,ind2;
	double guess1=-1,guess2=-1;
	
	ind1=index_to_solve[1];
	ind2=index_to_solve[2];
	
	if(ind1==1) { //solve tmax & ??
		guess1=sqrt(L*C);
		if(ind2==2) guess2=U0/R; //Imax
		if(ind2==3) guess2=Imax*R; //U0
		if(ind2==4) guess2=U0/Imax; //R
		if(ind2==5) { //L
			eqselection=fIsU0;
			Newton(ind2,1e-12);
			eqselection=fIsF0;
			Newton(ind1,1e-12);
			guess1=parameters[ind1];
			guess2=parameters[ind2];
		}
		if(ind2==6) { //C
			eqselection=fIsU0;
			Newton(ind2,1e-12);
			eqselection=fIsF0;
			Newton(ind1,1e-12);
			guess1=parameters[ind1];
			guess2=parameters[ind2];
		}
	}
	if(ind1==2) { //solve Imax & ??
		guess1=U0/R;
		if(ind2==3) { //U0
			guess1=0;
			guess2=0;
		}
		if(ind2==4) { //R
			eqselection=fIsF0;
			Newton(ind2,1e-9);
			guess1=U0/parameters[ind2];
			guess2=parameters[ind2];
		}
		if(ind2==5) guess2=pow(tmax,2)/C; //L
		if(ind2==6) guess2=pow(tmax,2)/L; //C
	}
	if(ind1==3) { //solve U0 & ??
		guess1=R*Imax;
		if(ind2==4) { //R
			eqselection=fIsF0;
			Newton(ind2,1e-9);
			guess1=parameters[ind2]*Imax;
			guess2=parameters[ind2];
		}
		if(ind2==5) guess2=pow(tmax,2)/C; //L
		if(ind2==6) guess2=pow(tmax,2)/L; //C
	}
	
	if(ind1==4) { //solve R & ??
		guess1=U0/Imax*2;
		if(ind2==5) guess2=pow(tmax,2)/C; //L
		if(ind2==6) guess2=pow(tmax,2)/L; //C
	}
	if(ind1==5) { //solve L & ??
		guess1=2*R*pow(Imax,2)*tmax/(pow(U0,2))*1.0/STORED2OHM;
		if(ind2==6) guess2=pow(tmax,2)/C; //C
	}
	if(guess1==0 || guess2==0) {
		message="initial condition not defined...\n";
		return EXIT_FAILURE;
	}
	guesses[1]=guess1;
	guesses[2]=guess2;
	
	return EXIT_SUCCESS;
}

int rlc::setRLCParameters(int index_to_solve[]) {
	parameters[index_to_solve[1]]=0;
	parameters[index_to_solve[2]]=0;
	tmax=parameters[1];
	Imax=parameters[2];
	U0=parameters[3];
	R=parameters[4];
	L=parameters[5];
	C=parameters[6];
	k=-2;
	message="";
	
	return EXIT_SUCCESS;
}

int rlc::solve(int index_to_solve[3])
{	
	int status=0;
	int iteration=0, max_iterations=5;
	double guesses[3];
	double decrease_factor=0.6;
	
	setRLCParameters(index_to_solve);
	
	if(setPhysicalGuesses(index_to_solve,guesses)==EXIT_FAILURE) return EXIT_FAILURE;
	do {
		status=solve_system(index_to_solve[1],index_to_solve[2],guesses[1],guesses[2]);
		iteration++;
		if(status<=-11) guesses[1] *= decrease_factor;
		if(status<=-12) guesses[2] *= decrease_factor;
		if(status==EXIT_FAILURE) {
			guesses[1] *= decrease_factor;
			guesses[2] *= decrease_factor;
		}
	} while((status<0) && (iteration<=max_iterations));
	
	if(status==EXIT_SUCCESS && iteration>=max_iterations) status=EXIT_FAILURE;
	
	printf("k=%g in %d loops\n",k,iteration);
	if(k<2) message.append("oscillating regime");
	if(k==2) message.append("critically damped regime");
	if(k>2) message.append("overdamped regime - Umax=R*Imax");
	if(k==-2) message.append("something bad happened...");
	
	return status;
}
